/* changes
 - 091222 (v121): adapt FMR and DMR format due to an update of FMR file version 7 (from 'ResolutionX' and 'ResolutionY' to 'NrOfColumns' and 'NrOfRows')
 - 090522 (v120): updated link for internal help (line 74)
 - 291021 (v120): 'apply to' files are printed to BV Log tab with empty lines, but this is required for plugin.
 - 121120 (v111): reset size of font in help fields to 8pt
 - 191219 (v111): added printing of parameters to BV Log tab
 - 141119 (v111): more restrictions to continue with applying VDMs in case of NaNs
 - 311019 (v111): fixed small bug where undistortion would be applied for option 0 (estimate only)
 - 181019 (v111): fixed small bug when using several apply files
 - 031019 (v111): removed 'the order does not matter'
 - 120718: 
 - June 2018: added 'order' for Gaussian spinbox and help; added one tab in *.ui
 - June-July 2017: div.
 - 230517: added buttons for batch version
 - 230517: changed 'bvqx' to 'bv'
 - 230517: vdmBrowse(), and add() do not invoke the plugin anymore, but BrowseFile()
 - 230517: added batchFileBrowse(), startProcessing(), readBatchFile()
 - 290517: added ProjectParameters(), selectOption(), continued readBatchFile()
 - 300517: added Experiment(), closeAllProjects() from anatabacus
 */


var bv = BrainVoyager;
var $ = bv.FindChild;
const OPTION_ESTIMATE               = 0;
const OPTION_APPLY                  = 1;
const OPTION_ESTIMATE_APPLY_SAME    = 2;
const OPTION_ESTIMATE_APPLY_ANY     = 3;


var o = new Object;
o.dialogFileName   = "COPE_v121.ui";
o.dialogAccessName = "PluginDialog";


// Called for the Browse buttons in the Select file for estimation manually tab (prev: Estimate VDM tab).
function select_input(index) {
    bv.SetPluginStringParameter("task", "select_fmr_or_dmr");
    bv.ExecutePlugin();
    var filename = bv.GetPluginStringParameter("filename"); // do not use BrowseFile here, because the plugin reads the nr of vols
    if (filename.length > 1) {
        var edit = $("filenameEdit"+index);
        var spinBox = $("volumeSpinBox"+index);
        edit.text = filename;
        spinBox.maximum = bv.GetPluginIntParameter("nr_of_volumes");
        spinBox.value = bv.GetPluginIntParameter("display_volume");
    }
}


o.initDlg = function() {
    
    var dlg = this.PluginDialog;

    // General options tab
    $("LoadFilenamesButton").clicked.connect(this, this.batchFileBrowse);
    $("startButton").clicked.connect(this, this.startProcessing);
    $("cancelButton").clicked.connect(this, this.close);
    
    // Estimate VDM tab
    $("browseButton0").clicked.connect(function() { select_input(0); });
    $("browseButton1").clicked.connect(function() { select_input(1); });
    $("estimateCloseButton").clicked.connect(this, this.close);
    $("estimateRunButton").clicked.connect(this, this.estimate);

    // Apply VDM tab
    $("vdmBrowseButton").clicked.connect(this, this.vdmBrowse);
    $("addButton").clicked.connect(this, this.add);
    $("checkButton").clicked.connect(this, this.check);
    $("applyCloseButton").clicked.connect(this, this.close);
    $("applyRunButton").clicked.connect(this, this.apply);

    bv.PrintToLog($("textBrowser").source);
    var sample_path = bv.PathToSampleData;
    var html_index = sample_path.replace("SampleData/", "Extensions/Plugins/CopePluginHelp/index.html");
    var html_index_file =  "file:///".concat(html_index);
    $("textBrowser").source = new QUrl(html_index_file);
    bv.PrintToLog($("textBrowser").source);
    
}


// Called for the Close buttons.
o.close = function() {
    this.PluginDialog.close();
}


// Called for the Run button on the Estimate VDM tab.
o.estimate = function() {
    // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
    var file0 = $("filenameEdit0").text;
    bv.OpenDocument(file0);

    // set parameters
    bv.SetPluginStringParameter("file0", file0);
    bv.SetPluginStringParameter("file1", $("filenameEdit1").text);
    bv.SetPluginIntParameter("volume0", $("volumeSpinBox0").value - 1);
    bv.SetPluginIntParameter("volume1", $("volumeSpinBox1").value - 1);
    bv.SetPluginIntParameter("rllr", $("estimateRllrRadioButton").checked);
    bv.SetPluginIntParameter("use_ic", $("seRadioButton").checked);
    bv.SetPluginIntParameter("use_ssd", $("ssdRadioButton").checked);
    //bv.SetPluginIntParameter("apply_to_input", $("applyToInput").checked);// obsolete in v1.1
    bv.SetPluginIntParameter("apply_to_input", $("rbtnEstimateAndApply").checked);
    //bv.SetPluginIntParameter("correct_motion", $("correctMotion").checked);
    bv.SetPluginIntParameter("save_derivative", $("saveDerivative").checked);
    bv.SetPluginIntParameter("fast_algorithm", $("linearAlgorithm").checked);
    bv.SetPluginIntParameter("order", $("spbOrderGauss").value);

    // call C++ function
    bv.SetPluginStringParameter("task", "estimate");
    bv.ExecutePlugin();
}

// Called for by the first tab "several files" option; including "estimate and apply" checkbox
o.estimateApplyToSameBatch = function(exp) {
    
    // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
    var projectParams, i;
    for (i=0; i < exp.noOfProjects; i++) {
    
        bv.PrintToLog("Processing project: " + (i+1));
        projectParams = exp.arrayOfProjectParams[i];
        bv.OpenDocument(projectParams.estimateFile0);
        bv.SetPluginStringParameter("file0", projectParams.estimateFile0);
        bv.SetPluginStringParameter("file1", projectParams.estimateFile1);
        bv.SetPluginIntParameter("volume0", projectParams.estimateFileVolume0);
        bv.SetPluginIntParameter("volume1", projectParams.estimateFileVolume1);
        bv.SetPluginIntParameter("rllr", $("estimateRllrRadioButton").checked);
        bv.SetPluginIntParameter("use_ic", $("seRadioButton").checked);
        bv.SetPluginIntParameter("use_ssd", $("ssdRadioButton").checked);
        bv.SetPluginIntParameter("apply_to_input", $("rbtnEstimateAndApply").checked);
        //bv.SetPluginIntParameter("correct_motion", $("correctMotion").checked);
        bv.SetPluginIntParameter("save_derivative", $("saveDerivative").checked);
        bv.SetPluginIntParameter("fast_algorithm", $("linearAlgorithm").checked);
        bv.SetPluginIntParameter("order", $("spbOrderGauss").value); //// CHECK!! -1? // 05-06-18
    
        // call C++ function
        bv.SetPluginStringParameter("task", "estimate");
        bv.ExecutePlugin();
        
        this.closeAllProjects();
    }
}

o.vdmBrowse = function() {

    var filename = bv.BrowseFile("Please select the voxel displacement map", "*_vdm.map");
    if (filename.length > 1)
        $("vdmFilenameEdit").text = filename;
}

o.batchFileBrowse = function() {
    var filename = bv.BrowseFile("Please select the batch processing text file", "*.txt");
    if (filename.length > 1)
        $("batchFileLabel").text = filename;
}

o.add = function() {

    var filename = bv.BrowseFile("Please select the FMR or DMR project", "*.fmr *.dmr");
    if (filename.length > 1)
        $("textEdit").append(filename);
}


o.check = function() {
    bv.SetPluginStringParameter("files", $("textEdit").plainText);
    bv.SetPluginStringParameter("task", "check");
    bv.ExecutePlugin();
}


o.apply = function() {
    
    BrainVoyager.PrintToLog("Starting to apply one or several VDMs to: ");
    files_string = $("textEdit").plainText;
    files = files_string.split("\n");
    if (files.length == 0) {
        BrainVoyager.PrintToLog("No files found.");
        return;
    }
    bv.OpenDocument(files[0]);
    bv.SetPluginIntParameter("rllr", $("estimateRllrRadioButton").checked);
    bv.SetPluginStringParameter("vdm", $("vdmFilenameEdit").text);
    bv.SetPluginStringParameter("files", files_string);
    bv.SetPluginStringParameter("task", "apply");
    bv.ExecutePlugin();
}

o.applyBatch = function(exp) {
  
    var projectParams;
    BrainVoyager.PrintToLog("Apply correction to batch:\n" + exp.toString());
    for (var i=0; i < exp.noOfProjects; i++) {

        bv.PrintToLog("Applying correction to project: " + (i+1));
        projectParams = exp.arrayOfProjectParams[i];
        files = projectParams.applyFiles;
        if (files.length == 0) {
            bv.PrintToLog("No files to apply to listed.");
            return;
        }
        if (projectParams.isnan != 1) {
            bv.OpenDocument(files[0].trim()); // 17-10-19 UIT TIJDELIJK
            bv.SetPluginIntParameter("rllr", $("estimateRllrRadioButton").checked);
            bv.SetPluginStringParameter("vdm", projectParams.vdmFile0);
          //  this.vdmFile0 = "";
         //   this.vdmFile1 = "";
            
            var res = projectParams.applyFiles.toString();
            res2 = res.replace(/(,)/gm,"");
            bv.SetPluginStringParameter("files", res2);
            bv.SetPluginStringParameter("task", "apply");
            bv.ExecutePlugin();
            this.closeAllProjects();
        } else {
            bv.PrintToLog("No voxel displacement map for this dataset; please re-estimate.");
        }
   
    }
}

o.estimateBatch = function(exp) {
    
    // open FMR/DMR -- the C++ code needs an open FMR/DMR to create .map files
    var projectParams;
    for (i=0; i < exp.noOfProjects; i++) {
        
        bv.PrintToLog("Estimating voxel displacement map for project: " + (i+1));
        projectParams = exp.arrayOfProjectParams[i];
        bv.OpenDocument(projectParams.estimateFile0);
        bv.SetPluginStringParameter("file0", projectParams.estimateFile0);
        bv.PrintToLog("File 1: " + projectParams.estimateFile0);
        bv.SetPluginStringParameter("file1", projectParams.estimateFile1);
        bv.PrintToLog("File 2: " + projectParams.estimateFile1);
        bv.SetPluginIntParameter("volume0", projectParams.estimateFileVolume0);
        bv.SetPluginIntParameter("volume1", projectParams.estimateFileVolume1);
        bv.SetPluginIntParameter("rllr", $("estimateRllrRadioButton").checked);
        bv.SetPluginIntParameter("use_ic", $("seRadioButton").checked);
        bv.SetPluginIntParameter("use_ssd", $("ssdRadioButton").checked);
        bv.SetPluginIntParameter("apply_to_input", (exp.option == OPTION_ESTIMATE_APPLY_SAME));// if apply to same: this will be performed straight away in plugin
        //bv.SetPluginIntParameter("correct_motion", $("correctMotion").checked);
        bv.SetPluginIntParameter("save_derivative", $("saveDerivative").checked);
        bv.SetPluginIntParameter("fast_algorithm", $("linearAlgorithm").checked);
        bv.SetPluginIntParameter("order", $("spbOrderGauss").value);
        
        // call C++ function
        bv.SetPluginStringParameter("task", "estimate");
        bv.ExecutePlugin();
        projectParams.vdmFile0 = bv.GetPluginStringParameter("vdm");
        projectParams.isnan = bv.GetPluginIntParameter("isnan");
        if (projectParams.isnan == 0) {
            bv.PrintToLog("Name of VDM: " + projectParams.vdmFile0); // fixed in plugin 12-07-18
        } else {
            projectParams.isnan = 1;
        }
        exp.arrayOfProjectParams[i] = projectParams; // now updated with VDM filename
        bv.PrintToLog("End of estimating voxel displacement map for project: " + (i+1));
        this.closeAllProjects();
    }
}

o.startProcessing = function() {
    
    // print settings to BV Log tab
    BrainVoyager.PrintToLog("Phase encoding direction: " + ($("estimateRllrRadioButton").checked ? "RL-LR" : "AP-PA"));
    BrainVoyager.PrintToLog("Intensity correction: " + ($("seRadioButton").checked ? "yes" : "no"));
    BrainVoyager.PrintToLog("Similarity measure: " + ($("ssdRadioButton").checked ? "SSD" : "NCC"));
    BrainVoyager.PrintToLog("Save derivative: " + ($("saveDerivative").checked ? "yes" : "no"));
    BrainVoyager.PrintToLog("Order of Gaussian: " + $("spbOrderGauss").value);
    
    if ($("rbtnSingleProject").checked) {
        if (($("rbtnEstimateAndApply").checked) || ($("rbtnApplyOnly").checked)) {
            this.estimate();
        } else if ($("rbtnApplyOnly").checked) {
            this.apply();
        }
    } else if ($("rbtnBatch").checked) {

        var exp = new Experiment();
        success = this.readBatchFile(exp);
        if (success) {
            if (exp.option == OPTION_ESTIMATE){
                 this.estimateBatch(exp);
            } else if (exp.option == OPTION_ESTIMATE_APPLY_SAME) {
                this.estimateApplyToSameBatch(exp);
            } else if (exp.option == OPTION_APPLY) {
                this.applyBatch(exp);
            } else if (exp.option == OPTION_ESTIMATE_APPLY_ANY) {
                this.estimateBatch(exp); // this should create VDM files
                this.applyBatch(exp); // now apply VDM files to EPIs from same session
            }
        }
     }
    this.close(); // this does not seem to work
}

o.readBatchFile = function(exp) {
    
    /* 
     general format of filenames file
     
     general parameters
     no of projects:
     option: 0/1/2/3 (=estimate, apply or both to same, or both to any in session)
     
     for each project:
     estimate files:
     [estimate file 0:]
     [estimate file 1:]
     volumes:
     [volume 0:]
     [volume 1:]
     vdm file: (“none” if option == 2)
     no of apply files:
     [run 0:]
     [run 1:]
     [run n:]
*/
    
    var projcounter, projfilecounter;
    var success = true;
    var dummy;
    var textfilename;
    if ($("batchFileLabel").text.length > 1) {
        textfilename = $("batchFileLabel").text;
    } else {
        textfilename = bv.BrowseFile("Please select the text file with filenames", "*.txt");
    }
    bv.PrintToLog("Start reading: " + textfilename);
    var name = String(textfilename); // circumvention of problem with implicit strings
    var filenamesfile = new QFile(name);
    filenamesfile.open(new QIODevice.OpenMode(QIODevice.ReadOnly));
    var textstr = new QTextStream(filenamesfile);
    var dummy = textstr.readLine();
    line = dummy.split(':');
    exp.noOfProjects = parseInt(line[1]);
    dummy = textstr.readLine();
    line = dummy.split(':');
    exp.option = parseInt(line[1]);
    dummy = textstr.readLine(); // empty line
    var experimentArray = new Array(exp.noOfProjects);
    var project;
    if ((exp.option == OPTION_ESTIMATE) || (exp.option == OPTION_APPLY) || (exp.option == OPTION_ESTIMATE_APPLY_SAME)  || (exp.option == OPTION_ESTIMATE_APPLY_ANY)) {
        
        for (projcounter = 0; projcounter < exp.noOfProjects; projcounter++) {
            
            project = new ProjectParameters();
            project.isnan = 0;
            if ((exp.option == OPTION_ESTIMATE) || (exp.option == OPTION_ESTIMATE_APPLY_SAME) || (exp.option == OPTION_ESTIMATE_APPLY_ANY)) { // estimate or estimate & apply to same
                
                dummy = textstr.readLine(); // line: "files for estimation:" //replaced 28-06-17
                project.estimateFile0 = textstr.readLine(); // estimate a voxel displacement map (VDM) from EPI and opposite phase encoded EPI file.
                project.estimateFile1 = textstr.readLine();
                dummy = textstr.readLine(); // volumes for estimation:
                project.estimateFileVolume0 = (parseInt(textstr.readLine() - 1));
                project.estimateFileVolume1 = (parseInt(textstr.readLine() - 1));

            }
            if ((exp.option == OPTION_APPLY) || (exp.option == OPTION_ESTIMATE_APPLY_ANY)) { // apply or estimate & apply to any
                dummy = textstr.readLine();//vdm file:
                project.vdmFile0 = textstr.readLine();
            /*    line=dummy.split(':');
                //project.vdmFile0 = textstr.readLine(); //vdm file:
                project.vdmFile0 = line[1].replace(/(^\s+|\s+$)/g, "");
            */  dummy = textstr.readLine(); //apply vdm file to:
                line=dummy.split(' ');
                var nrFiles = parseInt(line[line.length-1]);
                project.setNrOfApplyFiles(nrFiles);
                BrainVoyager.PrintToLog("Apply to the following " + nrFiles + ":");
                var distortedEPIfilesArray = new Array();
                for (var projfilecounter = 0; projfilecounter < project.nrOfApplyFiles; projfilecounter++) { // these are the runs that the VDM needs to be applied to
                    var filename = textstr.readLine();
                     distortedEPIfilesArray.push(String(filename + "\n"));
                    BrainVoyager.PrintToLog(filename);
                }
                project.applyFiles = distortedEPIfilesArray;// This makes the array with filenames collapse to a string, filenames separated by a comma
            } // end apply or estimate & apply to any
            experimentArray[projcounter] = project;
            dummy = textstr.readLine();
        }
        
    } else {
        success = false;
    }
    filenamesfile.close();
    exp.arrayOfProjectParams = experimentArray;
    return success;
}

function ProjectParameters() {
    
    this.estimateFile0 = "";
    this.estimateFile1 = "";
    this.estimateFileVolume0 = 0;
    this.estimateFileVolume1 = 0;
    this.vdmFile0 = "";
    this.vdmFile1 = "";
    this.nrOfApplyFiles = 0;
    this.applyFiles = new Array(); // 11-07-17
    
    this.setNrOfApplyFiles = function(nrFiles) {
        this.nrOfApplyFiles = nrFiles;
    }
    this.toString = function() {

        var txt = "Estimate file 0: " + this.estimateFile0 + "\n" +
        "Estimate file 1: " + this.estimateFile1 + "\n" +
        "Estimate volume 0: " + this.estimateFileVolume0 + "\n" +
        "Estimate volume 1: " + this.estimateFileVolume1 + "\n" +
        "VDM file 0: " + this.vdmFile0 + "\n" +
        "VDM file 1: " + this.vdmFile1 + "\n" +
        "Number of files to correct: " + this.nrOfApplyFiles + "\n";
        for (var i=0; i<this.nrOfApplyFiles; i++) {
            txt += this.applyFiles[i] + "\n";
        }
        return txt;
    }
}
    
function Experiment() {
    
    this.noOfProjects = 0;
    this.option = 0;
    this.arrayOfProjectParams = []; // array of ProjectParameters objects
    this.toString = function() {
        var txt = "Number of projects:" + this.noOfProjects + "\n" +
        "Distortion estimation/correction option: " + this.option + "\n";
        for (var i=0; i<this.noOfProjects; i++) {
            txt += "Project: " + (i+1) + "\n" + this.arrayOfProjectParams.toString();
        }
        return txt;
    }
}
    

o.selectOption = function(option) {

    switch (option) {
        case 0:
            return "estimate";
        case 1:
            return "apply";
        case 2:
            return "estimate & apply to same";
        case 3:
            return "estimate & apply to any in session";
        default:
            return "invalid option";
    }
}

o.closeAllProjects = function() { 
    
        var doc, i;
        var pn = bv.PrintToLog;
        bv.PrintToLog("Closing all projects...");
        var docs = bv.Documents;
        var n_docs = docs.Count;
        
        if (n_docs == 0)
            return;
    
        for (var i=n_docs; i>0; i--)
        {
            docs = bv.Documents;
            n_docs = docs.Count;
            doc = bv.ActiveDocument;
            doc.Close();
        }
}

o
